#include "navigation_pile.hpp"
#include <cmath> // defines ilogb
#include <vector> // defines std::vector
#include <queue> // defines std::queue
#include <algorithm> // defines std::copy
#include <cassert> // defines assert

#include "dsexceptions.h" // defines Overflow and Underflow
#include "debugging.c++" // defines show


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::branch_pointer 
navigation_pile<T, Container, Compare>::ancestor(
navigation_pile<T, Container, Compare>::branch_pointer i, 
navigation_pile<T, Container, Compare>::level delta) const {
  while (delta > 0) {
    i = i.parent;
    i--;
  }
  return i;
}


template <class T, class Container, class Compare>
bool
navigation_pile<T, Container, Compare>::empty() const {
 return (_n == 0);
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type
navigation_pile<T, Container, Compare>::size() const {
  return _n;
}


// OBS OBS N present beclause of compile error
template <class T, class Container, class Compare>
navigation_pile<T, Container, Compare>::navigation_pile(size_type N,
							const Compare& x, 
							const Container& Y
							) : _n(Y.size()) {
  size_type clogn = ceiling_log(size());
  std::vector<branch_pointer> V(clogn, 0);
  if (Y.begin() == Y.end()) {
    return;
  }

  branch_pointer node;
  branch_pointer node_parent;
  if (Y.size() == 1) {
    _last_leaf = new leaf<T>(Y[0]);
    return;
  }

  typename std::vector<T>::const_iterator i = Y.begin();
  leaf_pointer lj, lk;
  branch_pointer bj, bk;
  element ej, ek;
  size_type j;
  for (int half_size = Y.size() / 2; half_size != 0; half_size--) {
    node = new branch<T>();
    lj = new leaf<T>(*i);
    i++;
    lk = new leaf<T>(*i);
    i++;
    ej = lj->e;
    ek = lk->e;
    node->leaf_first_child = lj;
    node->leaf_second_child = lk;
    lj->parent = node;
    lk->parent = node;
    if (less(ej, ek)) {
      node->p = lk;
    }
    else {
      node->p = lj;
    }
    _last_leaf = lk;
    j = 0;    
    while (V[j] != 0) {
      bj = V[j];
      bk = node;
      lj = bj->p;
      lk = bk->p;
      ej = lj->e;
      ek = lk->e;
      node = new branch<T>();
      if (less(ej, ek)) {
	node->p = lk;
      }
      else {
	node->p = lj;
      }
      node->branch_first_child = bj;
      node->branch_second_child = bk;
      bj->parent = node;
      bk->parent = node;
      V[j] = 0;
      j++;
    }

    V[j] = node;    
  }

  _root = node;
  node = 0;
  if (Y.size() & 1) {  // Y.size() odd
    node = new branch<T>();
    lj = new leaf<T>(*i);
    node->leaf_first_child = lj;
    node->leaf_second_child = 0;
    lj->parent = node;
    node->p = lj;
    _last_leaf = lj;
  }
  
  j = 0; 
  while (j < V.size() - 1) {
    if (node == 0 && V[j] != 0) {
      node_parent = new branch<T>();
      node_parent->branch_first_child = V[j];
      node_parent->branch_second_child = 0;
      node_parent->p = V[j]->p;
      V[j]->parent = node_parent;
      node = node_parent;
    }
    else if (node != 0 && V[j] == 0) {
      node_parent = new branch<T>();
      node_parent->branch_first_child = node;
      node_parent->branch_second_child = 0;
      node_parent->p = node->p;
      node->parent = node_parent;
      node = node_parent;
    }
    else if (node != 0 && V[j] != 0 ) {
      node_parent = new branch<T>();
      node_parent->branch_first_child = V[j];
      node_parent->branch_second_child = node;
      V[j]->parent = node_parent;
      node->parent = node_parent;
      lj = V[j]->p;
      lk = node->p;
      ej = lj->e;
      ek = lk->e;      
      if (less(ej, ek)) {
	node_parent->p = lk;
      }
      else {
	node_parent->p = lj;
      } 
      node = node_parent;
    }
    j++;
  }

  if (node != 0) {
    _root = node;
  }
}


template <class T, class Container, class Compare>
template <class InputIterator>
navigation_pile<T, Container, Compare>:: navigation_pile(InputIterator first, 
							 InputIterator last, 
							 const Compare& x,
							 const Container& Y
							 ) : _n((last - first) + Y.size()) {
  Container A;
  A.insert(A.end(), Y.begin(), Y.end());
  A.insert(A.end(), first, last);
  size_type clogn = ceiling_log(size());
  std::vector<branch_pointer> V(clogn, 0);

  if (A.begin() == A.end()) {
    return;
  }

  branch_pointer node;
  branch_pointer node_parent;
  if (A.size() == 1) {
    _last_leaf = new leaf<T>(A[0]);
    return;
  }

  typename std::vector<T>::iterator i = A.begin();
  leaf_pointer lj, lk;
  branch_pointer bj, bk;
  element ej, ek;
  size_type j;
  for (int half_size = A.size() / 2; half_size != 0; half_size--) {
    node = new branch<T>();
    lj = new leaf<T>(*i);
    i++;
    lk = new leaf<T>(*i);
    i++;
    ej = lj->e;
    ek = lk->e;
    node->leaf_first_child = lj;
    node->leaf_second_child = lk;
    lj->parent = node;
    lk->parent = node;
    if (less(ej, ek)) {
      node->p = lk;
    }
    else {
      node->p = lj;
    }    
    _last_leaf = lk;
    j = 0;    
    while (V[j] != 0) {
      bj = V[j];
      bk = node;
      lj = bj->p;
      lk = bk->p;
      ej = lj->e;
      ek = lk->e;
      node = new branch<T>();
      if (less(ej, ek)) {
	node->p = lk;
      }
      else {
	node->p = lj;
      }
      node->branch_first_child = bj;
      node->branch_second_child = bk;
      bj->parent = node;
      bk->parent = node;
      V[j] = 0;
      j++;
    }

    V[j] = node;    
  }

  _root = node;
  node = 0;
  if (A.size() & 1) {  // A.size() odd
    node = new branch<T>();
    lj = new leaf<T>(*i);
    node->leaf_first_child = lj;
    node->leaf_second_child = 0;
    lj->parent = node;
    node->p = lj;
    _last_leaf = lj;
  }
  
  j = 0; 
  while (j < V.size() - 1) {
    if (node == 0 && V[j] != 0) {
      node_parent = new branch<T>();
      node_parent->branch_first_child = V[j];
      node_parent->branch_second_child = 0;
      node_parent->p = V[j]->p;      
      V[j]->parent = node_parent;
      node = node_parent;
    }
    else if (node != 0 && V[j] == 0) {
      node_parent = new branch<T>();
      node_parent->branch_first_child = node;
      node_parent->branch_second_child = 0;
      node_parent->p = node->p;
      node->parent = node_parent;
      node = node_parent;
    }
    else if (node != 0 && V[j] != 0 ) {
      node_parent = new branch<T>();
      node_parent->branch_first_child = V[j];
      node_parent->branch_second_child = node;
      V[j]->parent = node_parent;
      node->parent = node_parent;
      lj = V[j]->p;
      lk = node->p;
      ej = lj->e;
      ek = lk->e;
      if (less(ej, ek)) {
	node_parent->p = lk;
      }
      else {
	node_parent->p = lj;
      } 
      node = node_parent;
    }
    j++;
  }

  if (node != 0) {
    _root = node;
  }
}


template <class T, class Container, class Compare>
navigation_pile<T, Container, Compare>::~navigation_pile(){
  dispose_pile();
}



template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::dispose_pile1(branch_pointer t, size_type h)
{
  if (t->branch_first_child != 0 && h > 1) {
    dispose_pile1(t->branch_first_child, h - 1);
  }

  if (t->leaf_first_child != 0 && h == 1)  {
    dispose_pile2(t->leaf_first_child);
  }

  if (t->branch_second_child != 0 && h > 1) {
    dispose_pile1(t->branch_second_child, h - 1);
  }

  if (t->leaf_second_child != 0 && h == 1) {
    dispose_pile2(t->leaf_second_child);
  }
  delete(t);
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::dispose_pile2(leaf_pointer t)
{
  delete(t);
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::dispose_pile()
{
  if (_n == 0) return;
  if (_n == 1) {
    delete(_last_leaf);
    return;
  }
  dispose_pile1(_root, ceiling_log(_n));
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::check_pile1(branch_pointer t, size_type h)
{
  if (h > 0) {
    leaf_pointer tmp_l = t->p;
    element tmp_e = tmp_l->e;
    std::cout << "h " << h << std::endl;
    std::cout << "Branch position " << t << std::endl;
    std::cout << "Branch position parent " << t->parent << std::endl;
    std::cout << "Leaf position " << tmp_l << std::endl;
    std::cout << "Leaf element " << tmp_e << std::endl;
    std::cout << "element  " << (tmp_e  / 10000000) << std::endl;
  }

  if (t->branch_first_child != 0 && h > 1) {
    check_pile1(t->branch_first_child, h - 1);
  }

  if (t->leaf_first_child != 0 && h == 1)  {
    check_pile2(t->leaf_first_child);
  }

  if (t->branch_second_child != 0 && h > 1) {
    check_pile1(t->branch_second_child, h - 1);
  }

  if (t->leaf_second_child != 0 && h == 1) {
    check_pile2(t->leaf_second_child);
  }
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::check_pile2(leaf_pointer t)
{
  std::cout << "Leaf position " << t << std::endl;
  std::cout << "Branch position parent " << t->parent << std::endl;
  element tmp_e = t->e;
  std::cout << "element " << tmp_e << std::endl;
  std::cout << "element  " << (tmp_e  / 10000000) << std::endl;
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::check_pile()
{
  std::cout << "Begin check <<<<<<<<<<<<<<<<<<<<<<<<<  " << std::endl;
  if (_n == 0) return;
  if (_n == 1) {
    element tmp_e = _last_leaf->e;
    std::cout << "element " << tmp_e << std::endl;
    return;
  }
  std::cout << "ceiling_log(_n) " << ceiling_log(_n) << std::endl;
  check_pile1(_root, ceiling_log(_n) );
  std::cout << "End check <<<<<<<<<<<<<<<<<<<<<<<<<  " << std::endl;
}


template <class T, class Container, class Compare>
const typename navigation_pile<T, Container, Compare>::value_type& 
navigation_pile<T, Container, Compare>::top() const {
  if (_n == 1) {
    return _last_leaf->e;
  }
  return _root->p->e;
}

template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::push(
const navigation_pile<T, Container, Compare>::value_type& x
) {
  insert_new_element(x);
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::leaf_pointer K
) {
  leaf_pointer j = _last_leaf;
  branch_pointer i = j->parent;
  int h = 0;
  while (i != _root) {
    if (i->branch_first_child != j) break; 
    h++;
    j = i;
    i = i->parent;
  }
  
  branch_pointer node;
  branch_pointer node_parent;
  node = branch_alloc.allocate(1);
  node->leaf_first_child = K;
  node->p = K;
  while (h > 0) {
    node_parent = branch_alloc.allocate(1);
    node_parent->branch_first_child = node;
    node_parent->p = K;
    node->parent = node_parent;
    node = node_parent;
    h--;
  }
  
  node_parent = i;
  node_parent->branch_second_child = node;
  node->parent = node_parent;  
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::pop() {
  if(_n == 0) throw Underflow();
  if (_n == 1) {
    erase_last_element();
    return;
  }

  leaf_pointer m = _root->p;
  leaf_pointer ell = _last_leaf;
  leaf_pointer k;
  branch_pointer i = first_ancestor_with_two_children(_last_leaf);
  if (i == ell->parent) {
    k = i->leaf_first_child;
  }
  else {
    k = i->branch_first_child->p;
  }
  if (ell == m) {
    erase_last_element();
    update_full_path(_last_leaf);
  }
  else if (i->p == ell) {
    swap_leaves(k, ell);
    swap_leaves(k, m);
    erase_last_element();
    update_partial_path(ell, i);
    update_full_path(k);
  }
  else {
    swap_leaves(m, ell);
    erase_last_element();
    update_full_path(ell);
  }
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_full_path(
navigation_pile<T, Container, Compare>::leaf_pointer L
) {
  branch_pointer i = L->parent;
  if (i == 0) return;
  leaf_pointer first_leaf = i->leaf_first_child;
  leaf_pointer second_leaf = i->leaf_second_child;

  if (second_leaf != 0) {
    i->p = less(second_leaf->e, first_leaf->e) ? first_leaf : second_leaf;
  }
  else {
    i->p = first_leaf;
  }
  
  branch_pointer second_branch;
  leaf_pointer ell;
  leaf_pointer m;
  while (i != _root) {
    i = i->parent;
    second_branch = i->branch_second_child;
    ell = i->branch_first_child->p;
    if (second_branch != 0) {
      m = second_branch->p;
      i->p = less(m->e, ell->e) ? ell : m;
    }
    else { 
      i->p = ell;
    }
  }
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::leaf_pointer j,
navigation_pile<T, Container, Compare>::branch_pointer i
) {
  branch_pointer kappa = j->parent;
  while (i != kappa) {
    kappa->p = j;
    kappa = kappa->parent;
  }
}


template <class T, class Container, class Compare>
void 
navigation_pile<T, Container, Compare>::update_partial_path(
navigation_pile<T, Container, Compare>::branch_pointer j,
navigation_pile<T, Container, Compare>::branch_pointer i, 
navigation_pile<T, Container, Compare>::leaf_pointer K, 
navigation_pile<T, Container, Compare>::leaf_pointer L
) {
  leaf_pointer kappa;

  while (i != j) {
    kappa = j->p;
    if (kappa != L) return;
    j->p = K;
    j = j->parent;
  }
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::insert_new_element(
const navigation_pile<T, Container, Compare>::element& z
) {
  leaf_pointer lp;
  lp = new leaf<T>(z);

  if(_n == 0) {
    _last_leaf = lp;
    _n++;
    return;
  }

  branch_pointer root_node;
  if(_n == 1) {
    root_node = new branch<T>();
    root_node->leaf_first_child = _last_leaf;
    root_node->leaf_second_child = lp;
    _last_leaf->parent = root_node;
    lp->parent = root_node;
    root_node->p = less(_last_leaf->e, lp->e) ? lp : _last_leaf;
    _root = root_node;
    _last_leaf = lp;
    _n++;
    return;
  }

  branch_pointer q;
  q = _last_leaf->parent;
  if (q->leaf_second_child == 0) {
    q->leaf_second_child = lp;
    lp->parent = q;
    _last_leaf = lp;
    bottom_up_update_on_special_path(q, lp);
    _n++;
    return;
  }
  
  branch_pointer node;
  if (q == _root) {
    node = new branch<T>();
    node->leaf_first_child = lp;
    node->leaf_second_child = 0;
    lp->parent = node;
    node->p = lp;
    _last_leaf = lp;
    root_node = new branch<T>();
    root_node->branch_first_child = _root;
    root_node->branch_second_child = node;
    _root->parent = root_node;
    node->parent = root_node;
    root_node->p = less(_root->p->e, lp->e) ? lp : _root->p;
    _root = root_node;
    _n++;
    return;
  }

  size_type i = 0;
  q = q->parent;
  while (q->branch_second_child != 0 && q != _root) {
    i++;
    q = q->parent;
  }

  if (q == _root) i++;
  node = new branch<T>();
  node->leaf_first_child = lp;
  node->leaf_second_child = 0;
  node->p = lp;  
  lp->parent = node;
  branch_pointer node_parent;
  while (i > 0) { 
    node_parent = new branch<T>();
    node_parent->branch_first_child = node;
    node_parent->branch_second_child = 0;
    node_parent->p = node->p;
    node->parent = node_parent;
    node = node_parent;
    i--;
  }

  if (q == _root) {
    root_node = new branch<T>();
    root_node->branch_first_child = _root;
    root_node->branch_second_child = node;
    _root->parent = root_node;
    node->parent = root_node;
    root_node->p = less(_root->p->e, lp->e) ? lp : _root->p;
    _root = root_node;
  }
  else {
    node->parent = q;
    q->branch_second_child = node;
  }
  
  bottom_up_update_on_special_path(node->parent, lp);
  _last_leaf = lp;
  _n++;
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::erase_last_element() {
  if (_last_leaf->parent == 0) {
    delete _last_leaf;
    _last_leaf = 0;
  }
  else {
    branch_pointer q = _last_leaf->parent;
    if (q->leaf_second_child == _last_leaf) {
      delete q->leaf_second_child;
      _last_leaf = q->leaf_first_child;
      q->leaf_second_child = 0;
      if (_n == 2) {
	delete q;
	_last_leaf->parent = 0;
	_root = 0;
      }
    }
    else {
      delete q->leaf_first_child;
      size_type counter = 0;
      while (q->branch_second_child == 0) {
	branch_pointer p = q->parent;
	delete q;
	counter++;
	q = p;
      }
      q->leaf_second_child = 0;
      branch_pointer p = q->branch_first_child;
      if (q == _root) {
	delete q;
	_root = p;
	_root->parent = 0;
      }
      if (counter > 0 ) counter--;
      while (counter != 0) {
	p = p->branch_second_child;
	counter--;
      }
      _last_leaf = p->leaf_second_child;
    }
  }
  _n--;
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::branch_pointer
navigation_pile<T, Container, Compare>::first_ancestor_with_two_children(leaf_pointer L) {
  if (L->parent == 0) {
    return 0;
  }
  else {
    branch_pointer q = L->parent;
    leaf_pointer r = q->leaf_second_child;
    if (r == L) {
      return q;
    }
    else {
      while (q->branch_second_child == 0) {
	q = q->parent;
      }
      return q;
    }
  }
}
    


// Calculates power z of two
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type
navigation_pile<T, Container, Compare>::pow_two(
navigation_pile<T, Container, Compare>::level z
) const {
  return (1 << z);
  //return int(pow(2, z));
}


// Calculates ceils{ log_2 n }
template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::size_type 
navigation_pile<T, Container, Compare>::ceiling_log(
navigation_pile<T, Container, Compare>::size_type n
) const {
  if (n <= 0) return 0;
  size_type i = ilogb(n);
  return i + (pow_two(i) != n);
}


// Calculates ilogb but returning zero if i <= 0
template <class T, class Container, class Compare>
int
navigation_pile<T, Container, Compare>::ilogb_zero(
double i
) const {
  if (i <= 0) return 0;
  return ilogb(i);
}

template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::swap_leaves(
typename navigation_pile<T, Container, Compare>::leaf_pointer i, 
typename navigation_pile<T, Container, Compare>::leaf_pointer j
) {
  branch_pointer p = i->parent;
  branch_pointer q = j->parent;
  i->parent = q;
  j->parent = p;
  if (p == q) {
    i = p->leaf_first_child;
    j = p->leaf_second_child;
    p->leaf_first_child = j;
    p->leaf_second_child = i;
    if (j == _last_leaf) {
      _last_leaf = i;
    }
    else if (i == _last_leaf) {
      _last_leaf = j;
    }
    return;
  }
  if (i == p->leaf_first_child) {
    p->leaf_first_child = j;
  }
  else {
    p->leaf_second_child = j;
  }
  if (j == q->leaf_first_child) {
    q->leaf_first_child = i;
  }
  else {
    q->leaf_second_child = i;
  }
  if (j == _last_leaf) {
    _last_leaf = i;
  }
  else if (i == _last_leaf) {
    _last_leaf = j;
  }
}


template <class T, class Container, class Compare>
typename navigation_pile<T, Container, Compare>::branch_pointer 
navigation_pile<T, Container, Compare>::bottom_up_on_special_path(
navigation_pile<T, Container, Compare>::branch_pointer j, 
const navigation_pile<T, Container, Compare>::element& x
) {
  leaf_pointer ell;
  branch_pointer i = j->parent;
  while (i != _root) {
    ell = i->p;
    if (less(ell->e, x)) {
      j = i;
      i = i->parent;
      if (i == 0) return j;
    }
    else {
      return j;
    }
  }
  return j;   
}


template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::bottom_up_update_on_special_path(
navigation_pile<T, Container, Compare>::branch_pointer j, 
const navigation_pile<T, Container, Compare>::leaf_pointer k
) {
  leaf_pointer ell;
  while (j != 0) {
    ell = j->p;
    if (less(ell->e, k->e)) {
      j->p = k;
      j = j->parent;
    }
    else {
      return;
    }
  }
  return;   
}


#include <is_sorted.c++> // defines cphstl::is_sorted 

// Utility function used for testing
template <class T, class Container, class Compare>
void
navigation_pile<T, Container, Compare>::sort_pile(
size_type n, 
const Compare& x
) {
  less = x;
  Container t(n);
  std::copy(A.begin(), A.end(), t.begin());
  std::sort(t.begin(), t.end(), less);
  Container E(n);
  // use i-- insted of int i = 0 and n++ to make is_sorted work
  int i = n - 1;  
  while(size() > 0) {
    E[i] = top();
    pop();
    i--;
  }
  assert(cphstl::is_sorted(E.begin(), E.end(), less));
}
